/********************************************************
 * MemoryKeyerV2.0.5.ino compiled with Arduino 1.0.6 or 1.6.1
 * Arduino iambic CW memory keyer for ham radio usage
 * wb8nbs@gmail.com
 * Links to the latest code version will be at the end of:
 * https://wb8nbs.wordpress.com/2015/02/18/arduino-based-iambic-keyers/
 * https://wb8nbs.wordpress.com/2015/03/03/fun-with-direct-digital-synthesis/
 * 
 * Many thanks to KC4IFB, his iambic code feels exactly like a WB4VVF Accukeyer.
 * Richard Chapman KC4IFB "Build a Low-Cost Iambic Keyer Using Open-Source Hardware"
 * QEX - February, 2009 - code available on the ARRL web site
 * 
 * Circular Queue implemented with help from examples at:
 * http://embedjournal.com/implementing-circular-buffer-embedded-c/
 *
 * Modification log:
 * Jan 2014 Mod to eliminate speed pot. Close both paddles for at least three 
 *   seconds, beeps, will go into speed adjust mode; dot increases speed, dash decreases.
 *   Again close both paddles 3 secs, beeps, exits speed adjust mode. (note: removed)
 * Jan 2014 Add Straight Key mode. Enabled if Dash (ring) pin is closed
 *          at power on or reset.
 * Jan 2014 Add toneOn flag so it doesnt set tone on every iteration. tiny85 doesnt like that.
 * May 2014 Add startup routine to beep out a boot message. Note: removed in Dec.
 * May 2014 Changed morse table method, add serial to morse capability, fixed messages in flash
 * May 2014 Added stored (by compiler) messages and keyboard speed change
 * Nov 2014 Add PS2 support, remove serial. Reset with both paddles closed will set 13 WPM.
 * Dec 2014 move memories to eeprom so they can be programable from keyboard.
 *          Added a function button. Add serial back in, will send from either kbd.
 *          Change paddle speed set. Add restore default to eeprom on reset, Add memory program.
 *          Changed morse table again to smaller tree derived table after discussion at
 *          http://raronoff.wordpress.com/2010/12/16/morse-endecoder/
 *          Added morse to ascii table so ascii memories can be programmed from paddles
 *          Added four morse practice modes. Built a box to put it in.
 * Jan 2015 Cleanup doIambickey
 * Feb 2015 V2.0 Significant changes converting to fully asynchronous event loop
 *          All delay() and spin waits removed from the main loop path.
 *          Remove numbers only practice. Change eeprom order. Add change sidetone.
 *          V2.01 minor tuning space generation in doIambic.
 *          V2.02 Converted sidetone generation to DDS/PWM sine wave with help from
 *          http://ko7m.blogspot.com/2014/08/direct-digital-synthesis-dds.html
 *          http://www.analog.com/media/en/training-seminars/tutorials/MT-085.pdf
 *          V2.03 changed sine table to half step offset. Saves a byte in the table.
 *          V2.04 added crude wave shaping to front and rear of keyed sidetone. Method used
 *          is timed switching between three amplitude tables. Had to debounce straight key.
 *          V2.0.4.1 Minor bug fix to disable interrupt while changing sine tables
 * Mar 2015 V2.0.5 change AtoM table from int to char saves 68 bytes but created setup bug
 *          initializing digital pins. Change do loop to i<19, i<21 was killing tone.
 *          Fix bug in initByte, aaByte = aByte -= 0x20; did not work in 1.6.1 compile.
 ********************************************************/

#include <EEPROM.h>             // needed to save memories and speed setting
#include <avr/pgmspace.h>       // need to store data in PROGMEM
#include "AtoM.h"               // ASCII to Morse translation table
#include "MtoA.h"               // Morse to ASCII translation table
#include "Canned.h"             // Defaults: 4 messages and frequently changed parms
// Useful AVR macros for setting and resetting bits
#define cbi(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit))  // clear a bit
#define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))   // set a bit

// Sine wave amplitude lookup tables length 64, range is zero to 255
// 64 samples per period but using quarter wave symmetry, table is only 16 bytes
// three tables are used so wave amplitude can start and stop smoother
// tables are switched by changing pointers
// note: half step offset values used to get around a symmetry glitch
#define Bits 10                 // # of bits to shift phase accumulator 
#define MSBMask 0x0f            // turns off 2 MSBs in table lookup
// Full amplitude table
const unsigned char sineTable1_0[16] PROGMEM = {
  134, 147, 159, 171, 182, 193, 204, 213, 222, 230, 237, 243, 248, 251, 254, 255
};
// Three quarter amplitude table
const unsigned char sineTable3_4[16] PROGMEM = {
  133, 142, 151, 160, 169, 177, 185, 192, 198, 204, 209, 214, 217, 220, 222, 223
};
// Half amplitude table
const unsigned char sineTable1_2[16] PROGMEM = {
  131, 137, 143, 149, 155, 160, 166, 170, 175, 179, 182, 185, 187,189, 190, 191
};

// These must all be marked as volatile as they are used
// in the sine synthesizer interrupt service routine
volatile unsigned int tuningWord;        // added every interrupt 
volatile unsigned int phaseAccumulator;  // running total of phase offset
//volatile const unsigned char *sineTable = sineTable1_2; // ISR pointer to sine table
unsigned char volatile const *sineTable = sineTable1_2; // ISR pointer to sine table


// sine synthesizer reference frequency (sample rate) is 32 khz
const int refFrequency = 16000000 / 510; // Atmel data sheet chap. 18

// PS2 keyboard library from http://playground.arduino.cc/Main/PS2Keyboard
#include <PS2Keyboard.h>

// Pin definitions
const int DATAPIN = 4;          // defined keyboard data pin
const int IRQPIN =  3;          // defined keyboard clock pin
PS2Keyboard keyboard;           // invoke the PS2 library

const int DOTIN = 5;            // LOW -> dot paddle closed - Should be jack TIP
const int DASHIN = 6;           // LOW -> dash paddle closed - Should be jack RING
// don't use pin 13 for keyout it gets hit several times during boot up.
const int KEYOUT = 8;          // drives the 2N2222 that closes the key connection
// note using the ISR for DDS requires TONE to be on pin 11
const int TONE = 11;            // pin to output side tone

const int firstButton = 14;     // define first of four sequential analog pins for buttons
const int lastButton = firstButton + 3; // last of the four sequential buttons
const int FUNCTION = A4;        // Function button

// define four possible states of the iambic machine
const int IDLE = 0;             // doing nothing
const int DASH = 1;             // playing a dash
const int DOT = 2;              // playing a dot
const int DELAY = 3;            // in the dot-length delay between two dot/dashes

int dotLength;                  // length of a dot, the basic unit of timing
boolean dotVal;                 // value of the dot paddle this cycle of main loop
boolean dashVal;                // value of the dash paddle this cycle of main loop
int currElt = IDLE;             // state of what the keyer output is sending right now
int nextElt = IDLE;             // state the keyer will go into when current element ends 
int lastElt = IDLE;             // previous state of the keyer
unsigned long time;             // used to store the millis() reading
unsigned long currEltendtime;   // what time should current element finish (millis)
unsigned long T0, T1, T2;       // used to switch waveshape tables

int sideTone;                   // current side tone frequency
boolean inTone = false;         // records if in side tone set mode
boolean lastTone = false;       // records if side tone mode was exited
boolean rampUp = false;         // set up sloped beginning of side tone 
boolean rampDown = false;       // set up sloped ending of side tone

int wordsPerminute;             // speed setting in words per minute
boolean inSpeed = false;        // records if in speed set mode
boolean lastSpeed = false;      // used to determine if speedmode has exited

boolean toneOn = false;         // remembers that tone was turned on
boolean skMode = true;          // flag is unset if in Straight Key mode
boolean keyTherig = true;       // false if NOT keying rig (program, practice)
boolean keyTherigQ = true;      // false if NOT keying rig until Q is empty
boolean commandMode = false;    // set if in keyboard command mode

char asciiChar;                 // morse converted to ascii for printing
boolean enteringIdle =  true;   // records that first idle time was recorded
int inByte;                     // char input from either keyboard or paddles
boolean gotCharkey = false;     // true if char is waiting to process from keyboards
boolean gotCharpad = false;     // true if char is waiting to process from paddles
int charsOut = 0;               // counts charcters written to serial port
char aByte = 0;                 // ASCII retrived from Queue or message memory

int CPM = 0;                    // which button that triggered code proctice mode

int sendQ[MAXQ];                // transmit character circular message buffer
int head = 0;                   // next position to insert into Queue
int tail = 0;                   // next position to retrieve from Queue

/********************************************************
 * Arduino setup
 ********************************************************/
void setup() {
    Serial.begin(9600);           // set up for serial keyboard input and output

  // Set up sine synthesizer, timer2 to phase correct PWM 32kHz clock
  pinMode(TONE, OUTPUT);
  timer2Setup();                // for sine synthesized sidetone.

  keyboard.begin(DATAPIN, IRQPIN, PS2Keymap_French); // initialize the PS2 keyboard

  // Gratuitous pullup write to *all* I/O pins so none are floating.
  // note - Arduino PRO mini has eight analog pins but cant write to A6 or A7
  for (int i=0; i<=19; i++) {
    if(i != TONE) digitalWrite(i,HIGH);  // must skip the pin ISR uses
  }

  pinMode(KEYOUT, OUTPUT);      // set mode on the i/o pins
  digitalWrite(KEYOUT, LOW);    // initially the key output is open (low) 

  // NOTE paddle pins idle HIGH, goes LOW (grounded) when paddle is closed.
  pinMode(DOTIN, INPUT);   
  pinMode(DASHIN, INPUT);

  // check for request to default, if so copy canned messages into EEPROM
  // EEPROM organization is:
  // byte 0,1 INT sideTone frequency
  // byte 2   BYTE wordsPerminute
  // bytes 3-207  Copy of the four canned messages 51 bytes each
  if (!Button(FUNCTION)) {      // Func down on reset means restore defaults
    sideTone = DEFAULTTONE;     // set default freq from Canned.h
    sidetoneToeeprom();         // write default tone to eeprom
    wordsPerminute=DEFAULTSPEED;// set default speed from Canned.h
    wpmToeeprom();              // write default speed to eeprom

      for (int i=0; i<=3; i++) { // copy default PROGMEM strings to EEPROM
      for (int j=0; j<=50; j++) {
        EEPROM.write(i * 51 + j + 3, pgm_read_byte_near(&(cMessg[i][j])));
      }
    }
    Serial.println(F("Default completed"));
    beeps(LOWTONE);             // Announce default write is complete
    while (!Button(FUNCTION)) {
    }                           // wait for func to be released
  }                             // end of restore defaults to EEPROM

  // retrieve last saved sideTone frequency setting
  sideTone = EEPROM.read(0) | (EEPROM.read(1) << 8);   // retrieve saved speed setting
  tuningWord = 65536 * sideTone / refFrequency;        // set initial tone freq

  // retrieve last saved speed setting
  wordsPerminute = (int)EEPROM.read(2); // retrieve last saved speed setting
  dotLength = 1200 / wordsPerminute;    // initialize the element length

  // Check for request to enter Code Practice Mode
  for (int i=firstButton; i <= firstButton + 2; i++) {  // scan the memory buttons
    if (!Button(i)) {
      CPM = i;                  // mark to go into code practice
      beeps(LOWTONE);           // announce that caught a button
    }
    while(!Button(i)) {         // spin wait for button to be released
    }
    randomSeed(millis());       // add entropy
  }

  if (CPM != 0) {
    Serial.print(F("\nPractice Mode ")); 
    Serial.print(CPM - firstButton + 1);

    switch (CPM - firstButton) {
    case 0:   
      Serial.print(F(": Letters Only")); 
      break;
    case 1: 
      Serial.print(F(": Letters and Numbers")); 
      break;
    case 2:   
      Serial.print(F(": Letters and Numbers and Punctuation")); 
      break;
    }

    Serial.print(F(" at "));
    Serial.print(wordsPerminute);
    Serial.println(F(" WPM"));
  }

  // Check for request to enter tone frequency set mode (Memory 4 button)
  if (!Button(firstButton + 3)) {
    inTone = true;              // tone mode flag
    beeps(LOWTONE);
    while (!Button(firstButton + 3)) { // spin until button is released
    }
  }

  // Set flag to enter Straight Key mode if only dash pin is LOW at startup
  // Straight key uses a 2 conductor plug which will ground the ring
  // of the three conductor paddle jack,
  if (!digitalRead(DASHIN)) {
    skMode = false;
    beeps(LOWTONE);             // announce Straight Key mode
    time = millis() + 500;      // waste a half second
    while (millis() < time) {   // spin
    }
  }

  beeps(MIDTONE);               // announce we are ready for business
  Serial.println(F("\nKeyer is ready\n"));
}

/********************************************************
 * Arduino main loop
 ********************************************************/
void loop() {
  static int pressed;           // holds pin # of button that was pressed

  // check if code practice mode was requested
  // CPM has pin number of button else zero
  // aByte is zero when Queue is empty
  if ((CPM != 0) && (aByte == 0)) {
    codePractice(CPM);          // Q is empty, make five more characters
  }

  // check if memories need to be played or recorded
  // ATMega328 Arduino analog pins are numbers 14 - 19
  // the following routine depends on four pins being sequentially assigned
  // scan only if no button is pending and NOT in tone change mode
  if ((pressed == 0) && !inTone) {
    for (int i=firstButton; i <= lastButton; i++) { // scan the memory buttons

      if(!Button(i)) {
        pressed = i;            // caught a button down  

        if (!Button(FUNCTION)) {   // is the func button pressed?
          recordMessage (pressed); // call the message write routine
          pressed = 0;          // deactivate that button
        }
        break;                  // abort the scan
      }
    }
  }

  if (pressed) {                // any button pressed?
    if (Button(pressed)) {      // goes true when button released
      // insert into send queue the pin number of pressed button
      storeInBuffer((unsigned char)pressed); 
      pressed = 0;              // reset the button pressed
    }
  }

  // poll the two possible keyboards
  gotCharkey = false;
  if(Serial.available()) {      // Process a serial char if present
    inByte = Serial.read();     // get one byte
    gotCharkey = true;          // remember we have a character pending
  }

  if(keyboard.available()) {    // Process a PS2 char if present
    inByte = keyboard.read();   // get one byte          
    gotCharkey = true;          // remember we have a character pending
  }

  // If keyboard was read, do something with it
  if(gotCharkey) {
    processByte(inByte);        // check and Queue this ASCII character
  }

  time = millis();              // get the current time
  waveShape();                  // manage the sidetone generation

  // Straight Key Mode or Iambic Mode
  if (skMode) {
    doIambicKey();
  }
  else {
    doStraightKey();
  }
}                               // end of loop()

/********************************************************
 * Iambic keyer mode
 * Original code is from KC4IFB work published in QEX
 * is slightly modified to also send morse from a keyboard 
 * or memory, change sending speed, or sidetone frequency.
 * global keyTherig false blocks rig keying while programming 
 * memories or while in code practice mode.
 * GLobal keyTherigQ blocks keying until the Queue is empty.
 ********************************************************/
void doIambicKey() {

  static byte mByte = 0;        // holds an encoded morse byte being sent
  //  static char aByte = 0;        // ASCII retrived from Q or message
  static int aPos = 0;          // pointer to ASCII byte in message memory
  static unsigned long idleEntry; // records time that idle state was entered
  static unsigned long charSpacetime = time; // time for character space 
  static unsigned long wordSpacetime = time; // time for word space
  static boolean spaceNeeded = false;// records if an ascii space has been issued
  static byte bitMask = 0x01;   // used to assemble MtoA table index
  static int tabIndex = 0;      // assembled index into MtoA table

  // Poll the paddles *only* if no queued ASCII byte is pending
  // aByte is set to zero if the Queues are empty
  if (aByte == 0) {
    dotVal = digitalRead(DOTIN); 
    dashVal = digitalRead(DASHIN); 
  }

  // Check func button to see if speed mode requested
  if (!Button(FUNCTION)) {      // Func held down means set speed mode
    inSpeed = true;             // speed mode flag true means in set speed mode
    digitalWrite(KEYOUT,LOW);   // unkey the transmitter (and LED)
  } 
  else inSpeed = false;         // NOT in speed set mode

    // Preserve speed in eeprom but write only when speed mode is exited
  // lastSpeed is set only if speed was actually changed
  if (!inSpeed && lastSpeed) {  // NOT in speed mode now
    wpmToeeprom();              // writes Words Per Minute to eeprom
    lastSpeed = false;
  }

  // check Mem 4 button to see if tone mode needs to be exited
  if (inTone && !Button(lastButton)) {  // mem 4 pressed

    while (!Button(lastButton)) { // spin until mem 4 is released
    }

    //  When tone mode exits preserve sidetone freq in eeprom but write
    //  only if tone was actually changed (lastTone has been set).
    if (lastTone) {
      sidetoneToeeprom();
    }
    inTone = false;             // tone mode flag
    lastTone = false;
  }

  switch(currElt) {             // four cases based on current state

  case DASH:

    if ((dotVal == LOW) && (nextElt == IDLE)) {  // going to iambic mode 
      nextElt = DOT;
    }

    if (time >= currEltendtime) { // at end of current dash
      lastElt = DASH;
      currElt = DELAY;          // a delay will follow the dash 
      currEltendtime = time + dotLength;

      // if in sidetone adjust mode, dash paddle lowers frequency
      if (inTone) {
        changeFreq(-1);
        lastTone = true;
      }
      // if in speed adjust mode, dash paddle decreases speed
      else if (inSpeed) {
        changeSpeed( -1 );
        lastSpeed = true;
      }
    }

    // close keyer output and start sidetone while dash is sent
    if (!toneOn) {              // flag ensures tone is only turned on once
      if(!inTone && !inSpeed && keyTherig && keyTherigQ) {
        digitalWrite(KEYOUT,HIGH);
      }
      sineTone(TONE, sideTone); // used to be the Arduino tone()
      toneOn = true;            // remembers side tone is on and rig is keyed

      // add a dash bit into tabIndex for morse decode
      if (aByte == 0) {         // but only if sending from paddles
        // a dash adds in a zero (just shift mask left)
        if ((bitMask & 0x40) != 0x40) {  // no shift if at max length (7 bits)
          bitMask = bitMask << 1; // move mask to next higher bit position
        }
      }
    }
    break;                      // end of DASH case

  case DOT:

    // note if bitmask is not 0x01 then a character is in progress from paddles
    if ((dashVal == LOW) && (nextElt = IDLE)) {  // going to iambic mode
      nextElt = DASH;
    }

    if (time >= currEltendtime) { // at end of current dot 
      lastElt = DOT;            // a delay will follow the dot 
      currElt = DELAY; 
      currEltendtime = time + dotLength;

      // if in sidetone adjust mode, dot paddle raises frequency
      if (inTone) {
        changeFreq(+1);
        lastTone = true;
      }
      // if in speed adjust mode, dot paddle increases speed
      else if (inSpeed) {
        changeSpeed( +1 );
        lastSpeed = true;
      }
    }

    // close keyer output and start sidetone while dot is sent
    if (!toneOn) {              // flag ensures tone is only turned on once
      if(!inTone && !inSpeed && keyTherig && keyTherigQ) {
        digitalWrite(KEYOUT,HIGH);
      }
      sineTone(TONE, sideTone); // used to be the Arduino tone()
      toneOn = true;            // remembers that side tone is on and rig is keyed

      // add a dot (a one) into tabIndex for morse decode
      if (aByte == 0) {         // but only if sending from paddles
        tabIndex = tabIndex | bitMask;   // adds in a one
        if ((bitMask & 0x40) != 0x40) {  // no shift if at max length 7 bits
          bitMask = bitMask << 1; // move mask to next higher bit position
        }
      }
    }
    break;                      // end of DOT case

  case IDLE:                    // not sending, nor delay after a dot or dash

    // Are either (or both) of the paddles closed?
    if ((dotVal == LOW) && (dashVal == HIGH)) { // only dot closed 
      lastElt = IDLE; 
      currElt = DOT;            // set DOT mode
      currEltendtime = time + dotLength;
      return;
    }

    else if ((dotVal == HIGH) && (dashVal == LOW)) { // only dash closed
      lastElt = IDLE; 
      currElt = DASH;           // set DASH mode
      currEltendtime = time + (3 * dotLength);
      return;
    } 

    // *neither* paddle is closed
    // the morse character *may* be complete
    if(enteringIdle) {          // first time on this entry?
      idleEntry = time;         // mark entry time for space triggers
      enteringIdle = false;     // only mark the time once
      // time at which a character space will be triggered
      charSpacetime = idleEntry + dotLength * CSPACEWAIT;
      // time at which a word space will be triggered
      wordSpacetime = idleEntry + dotLength * WSPACEWAIT;
    }

    // don't bother printing assembled character or space if 
    // character was from queue or from memory message
    if (aByte == 0) {           // Char is in progress & Q is empty

      // delay before finishing an ascii decode from paddles
      // if paddles were active, bitMask will have been left shifted
      if((bitMask != 0x01) && (time > charSpacetime)) {  // from paddles

        // there is a morse character waiting and idle time has expired
        // finish assembling the morse decode index, lookup ASCII and print
        tabIndex = tabIndex | bitMask;  // or in the final stop bit
        asciiChar = (char)pgm_read_byte(&MtoA[tabIndex]); // get ASCII translation
        echoChar(inByte = (int)asciiChar); // print ASCII equiv of morse char

          // memory writes from paddles are lower case, else chars run together
        if (inByte >= 0x41) inByte += 0x20;
        gotCharpad = true;      // flag for possible eeprom write 
        bitMask = 0x01;         // re-initialize the index mask
        tabIndex = 0x00;        // and re-initialize the index
        spaceNeeded = true;     // a space character MAY be needed
      }

      // if a sufficiently long IDLE period has elapsed and the paddles
      // haven't started another character, print and store a word space
      if(spaceNeeded && (time > wordSpacetime)) {                                                                          
        echoChar(0x20);         // send a space to serial terminal
        inByte = 0x20;          // space for eeprom write subroutine
        gotCharpad = true;      // flag for eeprom write subroutine 
        spaceNeeded = false;    // record that one space was issued
      }
    } 

    // Done with previous character, check for pending action.
    // The following sorts out four possible states:
    // - We are in the middle of a button message, get next ASCII from EEPROM
    // - There is no memory message in progress and Q is empty
    // - Q not empty, and Q contains ASCII (>= 0x20)
    // - Q not empty, and Q contains a memory message pin number (< 0x20)
    if (aPos) {                 // aPos active, memory message in progress
      // get next ASCII byte from message memory until NULL
      aByte = EEPROM.read( ++aPos );

      if (aByte) {              // if not NULL, get morse byte
        mByte = initByte(aByte); // check and lookup morse
      } 
      else {                    // got null, end of message 
        aPos = 0;               // position of zero indicates not reading msg
        keyTherigQ = true;      // allow rig keying after speed or tone set
      }
      break;                    // skip queue check until msg is completed
    }

    if (head != tail) {         // Queue not empty, get new char started
      aByte = readFromBuffer(); // get next byte from circular Queue

      if (aByte >= 0x20) {      // >= space, byte must be ASCII
        mByte = initByte(aByte); // check and lookup morse
      }
      else {                    // Q contained a memory button indicator
        aPos = ((aByte - firstButton) * 51) + 3; // point to start of msg
        aByte = EEPROM.read(aPos); // get the first ASCII byte of the message
        mByte = initByte(aByte); // check and lookup morse
      }                         // end message button initialize
    }                           // end head != tail
    else { 
      aByte = 0;                // Signal that Queue was empty
      keyTherigQ = true;        // allow rig keying after speed or tone set
    }
    break;                      // end the IDLE case

  case DELAY:                   // Element length delay after dot or dash

    if (time >= currEltendtime) { // delay time expired?  
      // delay has expired, set a new current element and time
      currElt = nextElt;

      // check if we are playing from the queue
      if (mByte != 0) {         // not empty, pull next elt from morse byte

        if (mByte != 0x40) {    // not finished with this morse byte
          // mask and read bit six from morse byte
          if (mByte & 0x40) {   // need a dit next
            currElt = DOT;
          }
          else {                // need a dah next
            currElt = DASH;
          }
          mByte = (mByte << 1) & 0x7f;// lshift, truncate to right seven bits
        }                       // finished set morse bit

        else {                  // last element bit has been shifted out
          echoChar(aByte);      // print the ASCII from queue or msg memory
          // make an intercharacter space by extending DELAY two element times
          // but only if lower case, so Prosigns can be sent run together
          if((aByte < 0x41) || (aByte > 0x5A)) {
            currEltendtime = currEltendtime = time + (2 * dotLength);
            currElt = DELAY;
          }
          mByte = 0;            // marks that morse byte is done
          break;
        }
      }

      // set up the timeouts
      if (currElt == DOT) { 
        currEltendtime = time + dotLength; 
      }
      else if (currElt == DASH) { 
        currEltendtime = time + (3 * dotLength);
      }
      lastElt = DELAY; 
      nextElt = IDLE;           // default next element is IDLE
    }

    // during the delay, if either paddle is pressed, save it to play after the delay
    if ((lastElt == DOT) && (dashVal == LOW) && (nextElt == IDLE)) { 
      nextElt = DASH;           // override default IDLE
    }
    else if ((lastElt == DASH) && (dotVal == LOW) &&(nextElt == IDLE)) {
      nextElt = DOT;            // override default IDLE
    }

    // note there is always a DELAY before IDLE
    if(toneOn) {                // do these things only once on entry
      // key output is open during the delay
      digitalWrite(KEYOUT,LOW);
      nosineTone(TONE);         // used to be the Arduino noTone()
      toneOn = false;
      enteringIdle = true;      // NOT in idle mode yet
    }
    break;                      // end of case DELAY
  }                             // end of switch currElt
}                               // End of Iambic Keyer mode

/********************************************************
 * Converts an ASCII character originating from keyboard 
 * or one of the message memories to a morse encoded byte.
 * Returns morse encoded byte from ASCII to Morse table (AtoM.h).
 * Sets global current element and shifts for next bit read.
 * Subsequent element bits are shifted in and read in DELAY.
 ********************************************************/
byte initByte(byte aByte) {
  byte mByte;                  // holds an encoded morse byte being sent

  if(aByte >= 0x61) {          // ASCII byte is lower case
    aByte -= 0x20;             // shift lc to uc for table lookup
  }
  // ASCII 0x20 (space) is item zero in the table
  mByte = pgm_read_byte(&AtoM[aByte - 0x20]); // scale down

  // initialize current Element with bit six from new morse byte
  // further bit extracts and shifts are in DELAY
  if (mByte != 0x40) {         // is NOT a space
    if (mByte & 0x40) {        // need a dit next
      currElt = DOT; 
      currEltendtime = time + dotLength;
    }
    else {                      // need a dah next
      currElt = DASH; 
      currEltendtime = time + (3 * dotLength);
    }
    lastElt = IDLE;
    mByte = (mByte << 1) & 0x7f; // lshift, truncate to right seven bits
  } 
  else {                        // special for space = morse 0x40
    currElt = DELAY;
    currEltendtime = time + (3 * dotLength);
  }
  enteringIdle = true;          // NOT in idle mode
  return mByte;
}

/********************************************************
 * validate and scale an ASCII keyboard character for sending
 ********************************************************/
void processByte(int inByte) {

  // add line break to console screen if cr/lf received from keyboard
  if ((inByte == 0x0a) || (inByte == 0x0d)) {
    charsOut = MAXLINE;         // trigger a newline
  }

  if (inByte == 0x005c) {       // backslash received
    echoChar((char)inByte);     // echo the backslash
    commandMode = true;         // activate keyboard command mode
  }

  // ignore the incoming byte if out of range
  // note backslash *is* out of range
  if(((inByte >= 0x20) && (inByte <= 0x5a)) || ((inByte >= 0x61) && (inByte <= 0x7a))) { 
    if(commandMode) processCommand (inByte); // detour to command function
    else storeInBuffer(inByte); // add this character to the send queue
  }
}

/********************************************************
 * Enter command mode and process commands
 * Commands are a backslash followed by a single letter
 * only one command at a time is processed
 ********************************************************/
void processCommand(int inChar) {

  echoChar(inChar); 

  switch ( inChar) {

  case 'w':                     // force an EEPROM write
    wpmToeeprom();
    sidetoneToeeprom();
    inTone = false;
    lastTone = false;
    break;

  case '+':                     // ASCII plus sign (speed up)
    changeSpeed( +1 );
    storeInBuffer ('S');        // send something so user can judge speed
    keyTherigQ = false;         // but suppress rig keying until Q is empty
    break;

  case '=':                     // Same as plus sign but no shift
    changeSpeed( +1 );
    storeInBuffer ('S');        // send something so user can judge speed
    keyTherigQ = false;         // but suppress rig keying until Q is empty
    break;

  case '-':                     // ASCII minus sign (speed down)
    changeSpeed( -1 );
    storeInBuffer ('S');        // send something so user can judge speed
    keyTherigQ = false;         // but suppress rig keying until Q is empty
    break;

  case 'u':                     // "u" key (tone UP)
    changeFreq( +1 );
    storeInBuffer ('T');        // send something so user can judge frequency
    keyTherigQ = false;         // but suppress rig keying until Q is empty
    break;

  case 'd':                     // "d" key (tone DOWN)
    changeFreq( -1 );
    storeInBuffer ('T');        // send something so user can judge frequency
    keyTherigQ = false;         // but suppress rig keying until Q is empty
    break;

  case '1':                     // Send canned message 1
    storeInBuffer(firstButton);
    break;

  case '2':                     // send canned message 2
    storeInBuffer(firstButton + 1);
    break;

  case '3':                     // send canned message 3
    storeInBuffer(firstButton + 2);
    break;

  case '4':                     // send canned message 4
    storeInBuffer(firstButton + 3);
    break;

  default:                      // command char not recognized
    break;
  }                             // end switch

  commandMode = false;          // always cancel command mode
}

/***************************************************************
 * Place a byte of data into the ring buffer
 * will not overrun the buffer
 **************************************************************/
void storeInBuffer(unsigned char data)
{
  unsigned int next = (unsigned int)(head + 1) % MAXQ;
  if (next != tail)
  {
    sendQ[head] = data;
    head = next;
  }
}

/***************************************************************
 * Retrieve a byte of data from the ring buffer
 * returns -1 if the buffer was empty
 **************************************************************/
char readFromBuffer()
{
  // if the head isn't ahead of the tail, we don't have any characters
  if (head == tail) {
    return -1;                  // quit with an error
  }
  else {
    char data = sendQ[tail];
    tail = (unsigned int)(tail + 1) % MAXQ;
    return data;
  }
}

/********************************************************
 * Straight Key mode
 * Blindly follow the DOTIN lead (jack tip)
 ********************************************************/
void doStraightKey(void) {
  static unsigned long debounceTime;

  // Don't read the key again until the debounce period expires
  if (millis() < debounceTime) return;

  if(!digitalRead(DOTIN)) {     // Key is closed

    if(!toneOn) {               // flag ensures tone is only turned on once

      // turn on the output and the tone
      digitalWrite(KEYOUT,HIGH);// close the keyer output
      sineTone(TONE, sideTone);
      toneOn = true;            // remember that side tone is on and rig is keyed
      debounceTime = time + 5;
    }
  }
  else {                        // Key is open
    if(toneOn) {                // flag ensures tone is only turned off once

      // turn off tone and unkey output
      digitalWrite(KEYOUT,LOW); // unkey the rig
      nosineTone(TONE);
      toneOn = false;           // Remember that tone is off and rig is unkeyed
      debounceTime = time + 5;
    }
  }
}                               //  End of Straight Key mode

/********************************************************
 * change the morse sidetone frequency
 ********************************************************/
void changeFreq(int freqSign) {

  sideTone += (freqSign * sideTone) / 20;      // increment or decrement 5%

  if(sideTone > MAXTONE) sideTone = MAXTONE;   // range check max
  if(sideTone < MINTONE) sideTone = MINTONE;   // range check min
}

/********************************************************
 * change the morse sending speed
 ********************************************************/
void changeSpeed(int speedSign) {

  wordsPerminute += speedSign;               // increment or decrement speed

  if(wordsPerminute > MAXWPM) wordsPerminute = MAXWPM; // range check max
  if(wordsPerminute < MINWPM) wordsPerminute = MINWPM; // range check min

  dotLength = 1200 / wordsPerminute;
}

/********************************************************
 * update the stored sidetone frequency information in EEPROM
 ********************************************************/
void sidetoneToeeprom (void) {

  Serial.print(F("\n Writing sidetone ")) ; 
  Serial.print(sideTone);
  Serial.println(F(" Hz to eeprom"));

  // beep out the new tone frequency in morse 
  storeInBuffer(' ');           // seperate the output

  if (sideTone > 999) {         // need four digits?
    storeInBuffer(0x30 + (sideTone / 1000));   // thousands digit
  }
  if (sideTone > 99) {          // need three digits?
    storeInBuffer(0x30 + ((sideTone / 100) % 10)); // hundreds digit
  }
  if (sideTone > 9) {           // need two digits?
    storeInBuffer(0x30 + ((sideTone / 10) % 10));  // tens digit
  }
  storeInBuffer(0x30 + (sideTone % 10));       // ones digit
  storeInBuffer(' ');           // seperate the output
  charsOut = MAXLINE;           // reinit the echo counter
  keyTherigQ = false;           // suppress tx keying

  // sideTone is an INT, write two bytes
  EEPROM.write(0,(byte)(sideTone & 0xFF)); // record speed in EEPROM lo byte
  EEPROM.write(1,(byte)(sideTone >> 8));   // record speed in EEPROM hi byte
}

/********************************************************
 * update the stored speed information in EEPROM
 ********************************************************/
void wpmToeeprom (void) {

  Serial.print(F("\n Writing speed ")) ; 
  Serial.print(wordsPerminute);
  Serial.println(F(" WPM to eeprom"));

  // beep out the new speed in morse 
  storeInBuffer(' ');

  if (wordsPerminute > 9) {     // need two digits?
    storeInBuffer(48 + (wordsPerminute / 10)); // tens digit
  }
  storeInBuffer(48 + (wordsPerminute % 10));   // ones digit
  storeInBuffer(' ');           // seperate the output
  charsOut = MAXLINE;           // reinit the echo counter
  keyTherigQ = false;           // suppress tx keying

  // WordsPerminute is an int but its range only needs one byte
  EEPROM.write(2,(byte)(wordsPerminute));      // record speed in EEPROM lo byte
}

/********************************************************
 * echo character to serial port and start new line if > 80 characters sent
 ********************************************************/
void echoChar(int inByte) {
  if (++charsOut > MAXLINE) {   // trigger newline
    Serial.println();
    charsOut = 0;
  }
  Serial.print((char)inByte);
}

/********************************************************
 * function to read analog pins as debounced digital inputs 
 * works with 10k pull up and 0.01 or 0.1 ufd cap across switch
 * has 33 percent hysteresis.
 * Beware this will spin forever on an unterminated analog input
 ********************************************************/
boolean Button(int Bpin) {
  int level = 600;              // middle of dead zone

  while(level > 400 && level < 800) {// spin if in dead zone
    level = analogRead(Bpin);   // try again
  }

  if(level >= 800) {
    return HIGH;                // in upper third
  } 
  else {
    return LOW;                 // in lower third
  }
}

/********************************************************
 * get a message from the keyboard OR PADDLES and record it
 * to eeprom, is tripped when func and one of the memory 
 * buttons are both down then releases. Func alone to finish.
 ********************************************************/
void recordMessage (int whichButton) {

  // spin wait for memory button to be unpressed
  while (!Button(whichButton)) {  
  }

  // spin wait for function button to be unpressed
  while (!Button(FUNCTION)) {  
  }

  Serial.print (F("\n Recording Memory "));
  Serial.println(whichButton - firstButton + 1);

  charsOut = MAXLINE;           // will trigger a new line on serial console

  beeps(LOWTONE);               // announce entering program mode
  keyTherig = false;            // do NOT key the radio
  // calculate where this button's message starts in eeprom 
  int startAddress = (whichButton - firstButton) * 51 + 3;
  int memBase = startAddress;   // memBase will increment as chars are added

  // The following statement hijacks the main loop while recording memory messages
  while (Button(FUNCTION)) {    // get chars until func is pressed again

    time = millis();            // get the current time
    waveShape();                // manage the sidetone generation

    gotCharkey = false;         // will be set if character from keyboard
    gotCharpad = false;         // will be set if character from paddles

    // poll the serial keyboard
    if (Serial.available()) {   // process a serial char if present
      inByte = Serial.read();
      gotCharkey = true;
    }

    // poll the PS2 keyboard
    if(keyboard.available()) {  // Process a PS2 char if present
      inByte = keyboard.read(); 
      gotCharkey = true;
    }

    // poll the paddles, transmitter keying is supppressed
    doIambicKey();

    if (gotCharkey || gotCharpad) {  // there is something to do 

      if (gotCharkey) processByte(inByte); // beep out only if keyboard

      if ((inByte != 0x0a) && (inByte != 0x0d)) {  // ignore CR and LF
        EEPROM.write(memBase++, (byte)inByte);// write and increment
        // max memory string is 50 char plus NULL, stall if greater
        if(memBase > (startAddress + 50)) memBase = startAddress + 50; 
      }
    }
  }                             // end while()

  // here when func is pressed again, spin wait for func to be unpressed
  while (!Button(FUNCTION)) {  
  }

  EEPROM.write(memBase, NULL);  // complete the string in eeprom

  Serial.print(F("\n Memory "));
  Serial.print(whichButton - firstButton + 1);
  Serial.print(F(", "));
  Serial.print(memBase - startAddress);
  Serial.println(F(" characters written"));
  charsOut = MAXLINE;           // will trigger a new line on serial terminal
  beeps(LOWTONE);               // announce completion program mode
  keyTherig = true;             // enable keying the radio
}

/********************************************************
 * Message button 1, 2, or 3 was down when Arduino was reset
 * enter code practice mode. The only exit is to reset again.
 * Practice character groups are stuffed into the send queue.
 ********************************************************/
void codePractice (int inButton) {
  int ranChar;                  // a random char picked with rannum
  int outCount = 0;             // count chars placed in send queue
  static int groups = -1;       // count 5 char groups completed
  keyTherig = false;            // do NOT key the radio

  inButton -= firstButton;      // reference button to zero

  while (outCount++ < 5) {      // make five random chars

    switch (inButton) {

    case 0:                     // button 1 ltrs only
      ranChar = random (26);    // ASCII ltrs range
      storeInBuffer(ranChar + 0x61);
      break;

    case 1:                     // button 2 ltrs + figs
      ranChar = random(36);     // universe of ltrs+figs

      if (ranChar < 26) {       // letter range
        storeInBuffer(ranChar + 0x61);
      }
      else {                    // number range
        storeInBuffer(ranChar - 26 + 0x30);
      }
      break;

    case 2:                     // button 3 ltrs + figs + punct Oh My!
      // Many ASCII punctuation chars have no morse equivalent so we 
      // ignore the initial choice and use the Pinata Algorithm (flail 
      // around in the dark until you hit something good).
      while ((unsigned int)pgm_read_byte(&AtoM[ranChar = random(59)]) == 0x40) {
      }
      if(ranChar > 0x20) ranChar += 32; // convert upper case to lower
      storeInBuffer(ranChar + 0x20); 
      break;
    }
  }

  // finally stuff some (defined in Canned.h) delay time in the sendQueue
  for (int i=0; i < PRACPAUSE; i++) { // see Canned.h
    storeInBuffer(' ');         // store trailing spaces
  }

  if ((++groups % 5) == 0) {
    Serial.print(F("\ngroup "));
    pz(groups + 1);             // seperate into lines of five groups
    Serial.print("   ");
    charsOut = outCount = 0;
  }
}

/********************************************************
 * just to get leading zeros on a four digit integer
 ********************************************************/
void pz(int k) {
  if(k < 1000 ) Serial.print(F("0"));
  if(k < 100 ) Serial.print(F("0"));
  if(k < 10 ) Serial.print(F("0"));
  Serial.print(k);
}

/********************************************************
 * Announcement tones, 3 quick beeps
 ********************************************************/
void beeps(int freq) {

  for (int i=0; i <= 2; i++) {
    // must generate tone using timers, not tone()
    unsigned long onTime = millis() + 60;
    unsigned long offTime = onTime + 50;

    sineTone(TONE, freq);
    while ((time = millis()) < onTime) {
      waveShape();             // spin & waveshape
    }
    nosineTone(TONE);
    while ((time = millis()) < offTime) {
      waveShape();             // spin & waveshape
    }

    freq = (freq * 6) / 5;     // raise by a minor third
  }
}

/********************************************************
 * Set up to turn on the synthesized sine wave tone
 * This function replaces arduino tone()
 * The tone is actually turned on and off in waveShape
 ********************************************************/
void sineTone(int pin, int freq) {
  tuningWord = 65536 * freq / refFrequency;
  phaseAccumulator = 0xffff - tuningWord + 1; // sets zero phase to start
  rampUp = true;               // enable waveshaping for leading edge
  T0 = time;                   // turn on tone at half amplitude
  T1 = time + 1;               // 1 ms later to 3/4 amplitude
  T2 = time + 2;               // 1 ms later to full amplitude
}

/********************************************************
 * Set up to turn off the synthesized sine wave tone
 * This function replaces arduino noTone()
 * The tone is actually turned on and off in waveShape()
 ********************************************************/
void nosineTone(int pin) {
  // ramp down
  rampDown = true;              // enable waveshaping for trailing edge
  T0 = time;                    // initiate ramp down
  T1 = time + 1;                // 1 ms later down further
  T2 = time + 2;                // 1 ms later turn it off
}

/********************************************************
 * Output tone is turned on and off in this function with envelope
 * shaping for leading and trailing edges. Amplitude is
 * changed in three steps by switching between sine lookup tables.
 * This function must be called in the event loop to work.
 * Note rampDown leaves half amplitude table set during idle period.
 ********************************************************/
void waveShape() {

  // Shaping for tone leading edge envelope
  if (rampUp) {                 // Start of element tone generation
    if (time > T2) {            // Final switch to full amplitude
      TIMSK2 &= 0xFE;
      sineTable = sineTable1_0; // change table with interrupt disabled
      TIMSK2 |= 0x01;
      rampUp = false;           // done leading edge waveshaping
    }
    else if (time > T1) {       // 3/4 amplitude later at time 1
      TIMSK2 &= 0xFE;

      sineTable = sineTable3_4; // change table with interrupt disabled
      TIMSK2 |= 0x01;
      T1 = 0xffffffff;          // so it only does this once
    }
    else if (time > T0) {       // Turn on the wave at time zero       
      TIMSK2 |= 0x01;           // Enable timer 2 for sine synthesis
      T0 = 0xffffffff;          // so it only does this once
    }
  }

  // Shaping for tone trailing edge envelope
  if (rampDown) {               // End of element tone period
    if (time > T2) {            // Final time expired, turn off wave 
      TIMSK2 &= 0xFE;           // Disable timer 2 for sine synthesis
      OCR2A = 128;              // send steady PWM at half output voltage
      rampDown = false;         // done trailing edge waveshaping
    }
    else if (time > T1) {       // later switch to half amplitude
      TIMSK2 &= 0xFE;
      sineTable = sineTable1_2; // change table with interrupt disabled
      TIMSK2 |= 0x01;
      T1 = 0xffffffff;          // so it only does this once
    }
    else if (time > T0) {       // At time zero switch to 3/4 table
      TIMSK2 &= 0xFE;
      sineTable = sineTable3_4; // change table with interrupt disabled
      TIMSK2 |= 0x01;
      T0 = 0xffffffff;          // so it only does this once
    }
  }
}

/********************************************************
 * Setup timer2 with prescaler = 1, PWM mode = phase correct PWM
 * See the ATMega datasheet chap 18 for all the gory details
 ********************************************************/
void timer2Setup() {
  // Clock prescaler = 1
  sbi (TCCR2B, CS20);           // 001 = no prescaling
  cbi (TCCR2B, CS21);
  cbi (TCCR2B, CS22);

  // Phase Correct PWM
  cbi (TCCR2A, COM2A0);         // 10 = clear OC2A compare match
  sbi (TCCR2A, COM2A1);

  // Mode 1
  sbi (TCCR2A, WGM20);          // See table 18-8 in datasheet
  cbi (TCCR2A, WGM21);
  cbi (TCCR2B, WGM22);
}

/********************************************************
 * Timer 2 interrupt service routine (ISR) is used to generate
 * PWM for the sine wave DDS generator at 31.4kHz.
 * This version uses quarter wave symmetry to shrink table size by 4
 ********************************************************/
ISR(TIMER2_OVF_vect) {
  byte sineTableindex;          // calculated index into sine table     
  unsigned int whichHalf;       // for sine symmetry
  unsigned int whichQtr;        // for sine symmetry

  // Update phase accumulator and extract the sine table index from it
  phaseAccumulator += tuningWord;
  whichHalf = phaseAccumulator & 0x8000; // record which half of wave
  whichQtr =  phaseAccumulator & 0x4000; // record which quarter of wave

  // Right shift and mask because we only use the most significant bits
  // of the INT. Leave only enough bits for one quarter table range
  sineTableindex = (phaseAccumulator >> Bits) & MSBMask;

  // Look up current amplitude value for the sine wave being constructed.
  if(whichQtr) {                // in second or fourth quarter
    OCR2A = pgm_read_byte_near(sineTable + (MSBMask - sineTableindex));
  } 
  else {                        // in first or third quarter
    OCR2A = pgm_read_byte_near(sineTable + sineTableindex);
  }
  if(whichHalf) OCR2A = 255 - OCR2A; // invert the second half of wave
}



